@@ -22,6 +22,42 @@ abstract type AbstractObservable{T} end
 const addhandler_callbacks = []
 const removehandler_callbacks = []
 
+"""
+    mutable struct ObserverFunction <: Function
+
+Fields:
+
+    f::Function
+    observable::AbstractObservable
+    weak::Bool
+
+`ObserverFunction` is intended as the return value for `on` because
+we can remove the created closure from `obsfunc.observable`'s listener vectors when
+ObserverFunction goes out of scope - as long as the `weak` flag is set.
+If the `weak` flag is not set, nothing happens
+when the ObserverFunction goes out of scope and it can be safely ignored.
+It can still be useful because it is easier to call `off(obsfunc)` instead of `off(observable, f)`
+to release the connection later.
+"""
+mutable struct ObserverFunction <: Function
+    f::Any
+    observable::AbstractObservable
+    weak::Bool
+
+    function ObserverFunction(@nospecialize(f), @nospecialize(observable::AbstractObservable), weak::Bool)
+        obsfunc = new(f, observable, weak)
+        # If the weak flag is set, deregister the function f from the observable
+        # storing it in its listeners once the ObserverFunction is garbage collected.
+        # This should free all resources associated with f unless there
+        # is another reference to it somewhere else.
+        weak && finalizer(off, obsfunc)
+        return obsfunc
+    end
+end
+
+
+const OBSID_COUNTER = Base.Threads.Atomic{UInt64}(UInt64(0))
+
 """
     obs = Observable(val; ignore_equal_values=false)
     obs = Observable{T}(val; ignore_equal_values=false)
@@ -30,17 +66,18 @@ Like a `Ref`, but updates can be watched by adding a handler using [`on`](@ref)
 Set `ignore_equal_values=true` to not trigger an event for `observable[] = new_value` if `isequal(observable[], new_value)`.
 """
 mutable struct Observable{T} <: AbstractObservable{T}
-
     listeners::Vector{Pair{Int, Any}}
-    inputs::Vector{Any}  # for map!ed Observables
+    inputs::Vector{ObserverFunction}  # for map!ed Observables
     ignore_equal_values::Bool
+    id::UInt64
     val::T
-
     function Observable{T}(; ignore_equal_values::Bool=false) where {T}
-        return new{T}(Pair{Int, Any}[], [], ignore_equal_values)
+        Base.Threads.atomic_add!(OBSID_COUNTER, UInt64(1))
+        return new{T}(Pair{Int,Any}[], [], ignore_equal_values, OBSID_COUNTER[])
     end
     function Observable{T}(@nospecialize(val); ignore_equal_values::Bool=false) where {T}
-        return new{T}(Pair{Int, Any}[], [], ignore_equal_values, val)
+        Base.Threads.atomic_add!(OBSID_COUNTER, UInt64(1))
+        return new{T}(Pair{Int,Any}[], [], ignore_equal_values, OBSID_COUNTER[], val)
     end
 end
 
@@ -49,14 +86,14 @@ end
 
 Gets a unique id for an observable.
 """
-obsid(observable::Observable) = string(objectid(observable))
+obsid(observable::Observable) = string(getfield(observable, :id))
 obsid(obs::AbstractObservable) = obsid(observe(obs))
 
 function Base.getproperty(obs::Observable, field::Symbol)
     if field === :id
         return obsid(obs)
     else
-        getfield(obs, field)
+        return getfield(obs, field)
     end
 end
 
@@ -242,38 +279,6 @@ end
 
 Base.show(io::IO, ::MIME"application/prs.juno.inline", x::Observable) = x
 
-"""
-    mutable struct ObserverFunction <: Function
-
-Fields:
-
-    f::Function
-    observable::AbstractObservable
-    weak::Bool
-
-`ObserverFunction` is intended as the return value for `on` because
-we can remove the created closure from `obsfunc.observable`'s listener vectors when
-ObserverFunction goes out of scope - as long as the `weak` flag is set.
-If the `weak` flag is not set, nothing happens
-when the ObserverFunction goes out of scope and it can be safely ignored.
-It can still be useful because it is easier to call `off(obsfunc)` instead of `off(observable, f)`
-to release the connection later.
-"""
-mutable struct ObserverFunction <: Function
-    f::Any
-    observable::AbstractObservable
-    weak::Bool
-
-    function ObserverFunction(@nospecialize(f), @nospecialize(observable::AbstractObservable), weak::Bool)
-        obsfunc = new(f, observable, weak)
-        # If the weak flag is set, deregister the function f from the observable
-        # storing it in its listeners once the ObserverFunction is garbage collected.
-        # This should free all resources associated with f unless there
-        # is another reference to it somewhere else.
-        weak && finalizer(off, obsfunc)
-        return obsfunc
-    end
-end
 
 function Base.show(io::IO, obsf::ObserverFunction)
     io = IOContext(io, :compact => true)
@@ -452,7 +457,7 @@ function clear(@nospecialize(obs::Observable))
 end
 
 """
-    onany(f, args...)
+    onany(f, args...; weak::Bool = false, priority::Int = 0, update::Bool = false)
 
 Calls `f` on updates to any observable refs in `args`.
 `args` may contain any number of `Observable` objects.
@@ -461,7 +466,7 @@ All other objects in `args` are passed as-is.
 
 See also: [`on`](@ref).
 """
-function onany(f, args...; weak::Bool = false, priority::Int=0)
+function onany(f, args...; weak::Bool=false, priority::Int=0, update::Bool = false)
     callback = OnAny(f, args)
     obsfuncs = ObserverFunction[]
     for observable in args
@@ -470,6 +475,7 @@ function onany(f, args...; weak::Bool = false, priority::Int=0)
             push!(obsfuncs, obsfunc)
         end
     end
+    update && callback(nothing)
     return obsfuncs
 end
 
@@ -509,23 +515,23 @@ Observable{Number}(1.7320508075688772)
 
 can handle any number type for which `sqrt` is defined.
 """
-@inline function Base.map!(@nospecialize(f), result::AbstractObservable, os...; update::Bool=true)
+@inline function Base.map!(@nospecialize(f), result::AbstractObservable, os...; update::Bool=true, priority::Int = 0)
     # note: the @inline prevents de-specialization due to the splatting
     callback = MapCallback(f, result, os)
-    # appendinputs!(result, obsfuncs)
+    obsfuncs = ObserverFunction[]
     for o in os
-        o isa AbstractObservable && on(callback, o)
+        if o isa AbstractObservable
+            obsfunc = on(callback, o, priority = priority)
+            push!(obsfuncs, obsfunc)
+        end
     end
+    appendinputs!(result, obsfuncs)
     update && callback(nothing)
     return result
 end
 
 function appendinputs!(@nospecialize(observable), obsfuncs::Vector{ObserverFunction})  # latency: separating this from map! allows dropping the specialization on `f`
-    if !isdefined(observable, :inputs)
-        observable.inputs = obsfuncs
-    else
-        append!(observable.inputs, obsfuncs)
-    end
+    append!(observable.inputs, obsfuncs)
     return observable
 end
 
@@ -559,10 +565,10 @@ julia> map(length, obs)
 Observable(3)
 ```
 """
-@inline function Base.map(f::F, arg1::AbstractObservable, args...; ignore_equal_values=false) where F
+@inline function Base.map(f::F, arg1::AbstractObservable, args...; ignore_equal_values::Bool=false, priority::Int = 0) where F
     # note: the @inline prevents de-specialization due to the splatting
     obs = Observable(f(arg1[], map(to_value, args)...); ignore_equal_values=ignore_equal_values)
-    map!(f, obs, arg1, args...; update=false)
+    map!(f, obs, arg1, args...; update=false, priority = priority)
     return obs
 end
 
